import colorama
import collections
import StringIO

# colorama.init()


TerminalColor = collections.namedtuple('TerminalColor', ['fore', 'back'])


BLACK = TerminalColor(colorama.Fore.BLACK,  # pylint: disable=no-member
                      colorama.Back.BLACK)  # pylint: disable=no-member
RED = TerminalColor(colorama.Fore.RED,  # pylint: disable=no-member
                    colorama.Back.RED)  # pylint: disable=no-member
GREEN = TerminalColor(colorama.Fore.GREEN,  # pylint: disable=no-member
                      colorama.Back.GREEN)  # pylint: disable=no-member
YELLOW = TerminalColor(colorama.Fore.YELLOW,  # pylint: disable=no-member
                       colorama.Back.YELLOW)  # pylint: disable=no-member
BLUE = TerminalColor(colorama.Fore.BLUE,  # pylint: disable=no-member
                     colorama.Back.BLUE)  # pylint: disable=no-member
MAGENTA = TerminalColor(colorama.Fore.MAGENTA,  # pylint: disable=no-member
                        colorama.Back.MAGENTA)  # pylint: disable=no-member
CYAN = TerminalColor(colorama.Fore.CYAN,  # pylint: disable=no-member
                     colorama.Back.CYAN)  # pylint: disable=no-member
WHITE = TerminalColor(colorama.Fore.WHITE,  # pylint: disable=no-member
                      colorama.Back.WHITE)  # pylint: disable=no-member
RESET = TerminalColor(colorama.Fore.RESET,  # pylint: disable=no-member
                      colorama.Back.RESET)  # pylint: disable=no-member

NULL = TerminalColor('', '')

STYLE_DIM = colorama.Style.DIM  # pylint: disable=no-member
STYLE_NORMAL = colorama.Style.NORMAL  # pylint: disable=no-member
STYLE_BRIGHT = colorama.Style.BRIGHT  # pylint: disable=no-member
RESET_ALL = colorama.Style.RESET_ALL  # pylint: disable=no-member


def colorize(text, fore=NULL, back=NULL, style=STYLE_NORMAL):
    res = StringIO.StringIO()
    # Apply style
    res.write("".join([x or '' for x in [fore.fore, back.back,
                                         style]]))

    # Write message
    res.write(text)

    # Reset style
    res.write("".join([WHITE.fore, BLACK.back, STYLE_NORMAL]))
    return res.getvalue()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import collections
import logging


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def mock_namedtuple_class(tuple_class):
    class MockTuple(tuple_class):
        __EXCEPTION_SENTINEL = object()
        def __new__(cls, **kwargs):
            values = [kwargs.get(f, cls.__EXCEPTION_SENTINEL)
                      for f in tuple_class._fields]
            return tuple_class.__new__(cls, *values) #pylint: disable=star-args

        def __init__(self, **kwargs):
            for field in kwargs:
                if not field in self._fields:
                    raise ValueError("'{}' is not a valid field for {}".format(
                        field, tuple_class))
            kwargs = {f: kwargs.get(f, self.__EXCEPTION_SENTINEL)
                      for f in self._fields}
            tuple_class.__init__(self, **kwargs)

        def __getattribute__(self, attr):
            # Avoid recursion filtering _* lookups without doing self._* lookups
            value = tuple_class.__getattribute__(self, attr)
            if attr.startswith("_"):
                return value
            if value == self.__EXCEPTION_SENTINEL:
                raise AttributeError("Missing '{}' field in '{}' mock. (id={})"
                                     .format(attr, tuple_class.__name__,
                                             id(self)))
            return value
    return MockTuple


def mock_namedtuple(tuple_class, **kwargs):
    return mock_namedtuple_class(tuple_class)(**kwargs)


class _Sentinel(object):
    def __init__(self, parent, name):
        self._parent = parent
        self._name = name

    def __repr__(self):
        return "Sentinel({id}):{parent} @ {name}".format(
            id=hex(id(self)), parent=hex(id(self._parent)), name=self._name)


class Sentinels(object):
    def __init__(self):
        self._sentinels = {}

    def __getattribute__(self, attr):
        if attr.startswith('_'):
            return super(Sentinels, self).__getattribute__(attr)
        return self.__getitem__(attr)

    def __getitem__(self, key):
        try:
            return self._sentinels[key]
        except KeyError:
            sentinel = _Sentinel(self, key)
            self._sentinels[key] = sentinel
            return sentinel



from __future__ import absolute_import

from argparse import ArgumentParser
import logging
import os

from . import terminal, curses
from .tab import parse_tablature, transpose_tablature
from .instrument import UKELELE, Instrument


def _get_arg_parser():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest='subparser')
    add_parsers(subparsers)
    return parser


def add_parsers(subparsers):
    parser = subparsers.add_parser("parse")
    parser.add_argument('filename',
                        help='file to parse tablature from')
    parser.add_argument('-i', '--instrument', action='store', default=None,
                        help='Instrument to fetch chords for')
    parser.add_argument('-t', '--transpose', action='store', type=int,
                        default=0, help=('Number of steps to transpose the tab.'
                                         ' Defaults to 0'))
    parser.add_argument('-c', '--capo', action='store', type=int, default=0,
                        help='Capo position for the instrument. Defaults to 0')
    parser.add_argument("--curses",
                        action='store_true', dest='use_curses', default=False,
                        help="Use curses to show tablature")
    parser.add_argument("--debug",
                        action="store_true", dest="debug", default=False,
                        help="Sets logging level to DEBUG")
    parser.set_defaults(func=_parse_tablature)


def _parse_tablature(filename, instrument, use_curses, transpose, capo):
    with open(filename) as fin:
        lines = fin.readlines()
    tablature = parse_tablature(lines)
    if transpose:
        tablature = transpose_tablature(tablature, transpose)
    instrument = Instrument.from_name(instrument, UKELELE)
    if capo:
        instrument = instrument.capo(capo)
    if use_curses:
        curses.render_tablature(tablature, instrument)
    else:
        terminal.render_tablature(tablature, instrument)


def _extract_from_options(key, options):
    res = getattr(options, key)
    delattr(options, key)
    return res


def main():
    parser = _get_arg_parser()
    options = parser.parse_args()

    logging.basicConfig(filename='chorddb.log', filemode="w",
                        level=(logging.DEBUG
                               if _extract_from_options("debug", options)
                               else logging.INFO))
    subparser = _extract_from_options("subparser", options)

    try:
        func = _extract_from_options("func", options)
    except AttributeError:
        raise Exception("Missing 'func' attribute. Maybe subparser '{}' didnt "
                        "correctly set_defaults()?".format(subparser))

    func(**options.__dict__)

if __name__ == "__main__":
    main()



'''
Created on May 10, 2014

@author: ignacio
'''
import re
from ..notes import Key, KEY_RE
from ..utils.decorators import memoize
from ..utils.regexp import strict, re_search
from .variations import VARIATIONS_RE, VARIATIONS, VARIATIONS_NOTES


CHORD_RE = "{}(?:{})?(?:/{})?".format(KEY_RE, VARIATIONS_RE, KEY_RE)
_CAPTURING_CHORD_RE = "(({})({})?(?:/({}))?)".format(KEY_RE, VARIATIONS_RE,
                                                     KEY_RE)


class Chord(object):

    def __init__(self, key, variation=None, bass=None):
        self._key = key
        self._variation = variation
        self._bass = bass or self._key

    @property
    def key(self):
        return self._key

    @property
    def variation(self):
        return self._variation

    @property
    def bass(self):
        return self._bass

    def text(self):
        bass = ('' if self._key == self._bass
                else "/{}".format(self._bass.text()))
        return "{}{}{}".format(self._key.text(),
                               self._variation if self._variation else "",
                               bass)

    def bassless(self):
        ''' The same chord, but without bass '''
        return Chord(self._key, self._variation)

    def __str__(self):
        return "Chord:{}".format(self.text())

    def transpose(self, interval):
        return Chord(self._key.transpose(interval), self._variation)

    def __hash__(self):
        return hash((self._key, self._variation, self.bass))

    def __eq__(self, ochord):
        return (isinstance(ochord, Chord) and
                self.key == ochord.key and
                self.variation == ochord.variation and
                self.bass == ochord.bass)

    def __ne__(self, ochord):
        return not self == ochord

    @classmethod
    def parse(cls, text):
        mobj = re_search(strict(_CAPTURING_CHORD_RE), text)
        key, variation, bass_key = mobj.groups()[1:4]
        return cls(Key.parse(key), variation,
                   Key.parse(bass_key) if bass_key else None)

    @classmethod
    def extract_chordpos(cls, line):
        index = 0
        remainder = line
        res = []
        while True:
            mobj = re.search(_CAPTURING_CHORD_RE, remainder)
            if not mobj:
                break
            chord = mobj.group(1)
            chordpos = remainder.index(chord)
            res.append((Chord.parse(chord), chordpos + index))
            movement = chordpos + len(chord)
            remainder = remainder[movement:]
            index += movement
        return res

    @classmethod
    @memoize
    def all(cls):
        for key in Key.all():
            yield Chord(key)
            for variation in VARIATIONS:
                yield Chord(key, variation)



'''
Created on May 17, 2014

@author: ignacio
'''
from ..notes import Key


def map_variations_to_intervals(variations):
    return[
        [Key.parse(note).ord() for note in variation]
        for variation in variations]



_VARIATIONS_NOTES = {
    None: [["A", "C#", "E"]],
    "7": [["A", "C#", "E", "G"]],
    "m": [["A", "C", "E"]],
    "m7": [["A", "C", "E", "G"]],
    "maj7": [["A", "C#", "E", "G#"]],
    "dim": [["A", "C", "Eb"]],
    "aug": [["A", "C#", "F"]],
    "sus4": [["A", "D", "E"]],
    "m7b5": [["A", "C", "Eb", "G"]],
}
VARIATIONS_NOTES = {v: map_variations_to_intervals(nss)
                    for (v, nss) in _VARIATIONS_NOTES.items()}

VARIATIONS = sorted((v for v in _VARIATIONS_NOTES if v), reverse=True)
VARIATIONS_RE = "|".join(VARIATIONS)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import curses
import logging

from .pad import CursesPad


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class CursesScreen(object):
    def __init__(self, screen):
        self._screen = screen

    def get_dimensions(self):
        size_y, size_x = self._screen.getmaxyx()
        return size_x, size_y

    @property
    def width(self):
        return self.get_dimensions()[0]

    @property
    def height(self):
        return self.get_dimensions()[1]

    def write(self, xpos, ypos, text, color, attr=0):
        if not 0 <= ypos < self.height:
            logger.warning("Writing outside screen: ypos=%s not in [0,%s)]",
                           ypos, self.height)
            return
        if xpos >= self.width:
            logger.warning("Writing outside screen: xpos=%s > width=%s",
                           xpos, self.width)
            return
        if xpos < 0:
            logger.warning("Writing outside screen: xpos=%s < 0", xpos)
            text = text[abs(xpos):]
            xpos = 0
        if xpos + len(text) > self.width:
            logger.warning("Writing outside screen: text '%s' is too large "
                           "(len=%s + xpos=%s > width=%s)", text, len(text),
                           xpos, self.width)
            text = text[:self.width - xpos]
        self._screen.addstr(ypos, xpos, text, curses.color_pair(color) | attr)

    def write_all(self, blits):
        for blit in blits:
            self.write(blit.xpos, blit.ypos, blit.text, blit.color, blit.attr)

    def clear(self):
        self._screen.clear()

    def refresh(self):
        self._screen.refresh()

    def get_input_key(self):
        return self._screen.getch()

    def subpad(self, xpos, ypos, width, height, display_width=None,
               display_height=None):
        display_width = display_width or width
        display_height = display_height or height
        if xpos < 0:
            raise ValueError("xpos must be non-negative")
        if ypos < 0:
            raise ValueError("ypos must be non-negative")
        if width <= 0:
            raise ValueError("width must be positive")
        if height <= 0:
            raise ValueError("height must be positive")
        if xpos + display_width > self.width:
            logger.warning("Creating out of bounds pad: xpos=%s + "
                           "display_width=%s > %s", xpos, display_width,
                           self.width)
            display_width -= self.width - xpos
        if ypos + display_height > self.height:
            logger.warning("Creating out of bounds pad: xpos=%s + "
                           "display_height=%s > %s", xpos, display_height,
                           self.height)
            display_height -= self.height - ypos

        return CursesPad(
            # Extra 1 width, because fuck you curses, thats why
            curses.newpad(height, width + 1),
            xpos,
            ypos,
            width,
            height,
            display_width,
            display_height,
        )



from functools import wraps
import copy


def memoize(func):
    _cache = {}

    @wraps(func)
    def wrapper(*args, **kwargs):
        key = (args, tuple(sorted(kwargs.items())))
        try:
            return copy.copy(_cache[key])
        except TypeError:  # Unhashable key
            return func(*args, **kwargs)
        except KeyError:
            value = func(*args, **kwargs)
            _cache[key] = value
            return copy.copy(value)

    return wrapper



'''
Created on May 16, 2014

@author: ignacio
'''
import re


def strict(regexp):
    if regexp.startswith("^") or regexp.endswith("$"):
        raise ValueError("Regexp '{}' is already (semi)strict".format(regexp))
    return "^{}$".format(regexp)


class NoMatchError(ValueError):

    def __init__(self, pattern, string):
        ValueError.__init__(self, "Could not match pattern:'{}' "
                            "to string:'{}'".format(pattern, string))
        self._pattern = pattern
        self._string = string

    @property
    def pattern(self):
        return self._pattern

    @property
    def string(self):
        return self._string


def re_search(pattern, string, flags=0):
    mobj = re.search(pattern, string, flags=flags)
    if not mobj:
        raise NoMatchError(pattern, string)
    return mobj



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import collections
import logging


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


Tablature = collections.namedtuple('Tablature', ['lines'])
TabLine = collections.namedtuple('TabLine', ['type', 'data', 'original'])
ChordLineData = collections.namedtuple('ChordLineData', ['chords'])
LyricLineData = collections.namedtuple('LyricLineData', ['lyrics'])
TabChord = collections.namedtuple('TabChord', ['position', 'chord'])



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging

from nose.tools import eq_

from chorddb.instrument import Instrument
from chorddb.notes import KeyOctave


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def basic_capo_test():
    """ Basic check for `Instrument.capo` """
    capo_position = 2
    keyoctaves = [KeyOctave.parse(x + "0") for x in "ABCDEFG"]
    transposed = [ko.transpose(capo_position) for ko in keyoctaves]
    with_bass = Instrument("test", keyoctaves, 30, True)
    without_bass = Instrument("test", keyoctaves, 30, False)
    for original in [with_bass, without_bass]:
        capoed = original.capo(capo_position)
        eq_(capoed.has_bass, original.has_bass)
        eq_(capoed.keyoctaves, transposed)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging
from StringIO import StringIO


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name



"""
    3
 ||-o-|---|---|---|
 ||-o-|---|---|---|
 ||---|-o-|---|---|
 ||---|---|-o-|---|
 ||---|---|-o-|---|
 ||-o-|---|---|---|

"""


def draw_chord(fingering, reverse=True):
    positions = fingering.positions
    try:
        start = min(p for p in positions if p)
    except ValueError:
        start = 1

    if start <= 2:
        start = 1

    end = max(p for p in positions) - start
    if end < 3:
        end = 3

    lines = []

    lines.append("         {}".format(start))

    keyoctaves = [ko.key for ko in fingering.instrument.keyoctaves]

    values = zip(fingering.full_positions(), keyoctaves)
    if reverse:
        values.reverse()

    for position, keyoctave in values:
        line = StringIO()
        char = 'o' if position == 0 else 'x' if position == 'x' else ' '
        line.write(_line_start(char, keyoctave))
        for fret in xrange(0, end + 1):
            line.write(_fret(position == fret + start))
        lines.append(line.getvalue())

    return lines


def _line_start(char, keyoctave):
    return " {:3} {}||".format(keyoctave.text(), char)


def _fret(touched):
    return "-{}-|".format("o" if touched else "-")



'''
Created on May 10, 2014

@author: ignacio
'''
import collections
import logging
import re
import StringIO

from .parser import parse_tablature


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def transpose_line(line, interval):
    if line.type == 'chord':
        new_chords = [c._replace(chord=c.chord.transpose(interval))
                      for c in line.data.chords]
        return line._replace(data=line.data._replace(chords=new_chords))
    else:
        return line


def transpose_tablature(tablature, interval):
    new_lines = [transpose_line(line, interval) for line in tablature.lines]
    return tablature._replace(lines=new_lines)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import collections
import curses
import logging

from .color import DEFAULT_COLOR, CHORD_COLOR, FINGERING_COLOR, init_colors
from .state import RenderState
from .chord_drawer import draw_chord


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


CursesBlit = collections.namedtuple('CursesBlit', ['xpos', 'ypos', 'text',
                                                   'color', 'attr'])


def render_tablature(tablature, instrument):
    from . import wrapper
    wrapper(lambda s: _render(s, tablature, instrument))


def _render(screen, tablature, instrument):
    init_colors()
    chordpad_width, chordpad_height = _get_chordpad_dimensions(screen)
    vertical_division = screen.width - chordpad_width

    lyrics_pad = screen.subpad(
        0, 0,
        vertical_division - 1, len(tablature.lines),
        display_height=screen.height
    )
    chord_pad = screen.subpad(
        vertical_division + 1, 0,  # + 1 for left margin
        chordpad_width - 1, chordpad_height,  # -1 because of left margin
    )
    instructions_pad = screen.subpad(
        vertical_division + 1, chordpad_height + 2,  # +1/+2 for left/top margin
        chordpad_width - 1,  # -1 because of top margin
        screen.height - chordpad_height - 2,  # -2 because of left margin
    )

    state = RenderState.initial_state(tablature, instrument, screen)
    while state.running:
        screen.write_all(_get_division_blits(vertical_division, chordpad_width,
                                             chordpad_height, state))
        _update_lyrics_pad(lyrics_pad, state)
        _update_instructions_pad(instructions_pad, state)
        _update_chord_pad(chord_pad, state)
        screen.refresh()
        lyrics_pad.refresh(state.lyrics_position)
        instructions_pad.refresh()
        chord_pad.refresh()
        key = screen.get_input_key()
        logging.debug("Got key: '%s'", key)
        state = _process_input_key(key, state)
    logging.info("Im freeeee")


def _get_chordpad_dimensions(screen):
    width = min(36, screen.width / 2)
    height = min(15, screen.height / 2)
    return width, height


def _get_division_blits(vertical_division, chordpad_width, chordpad_height,
                        state):
    for ypos in xrange(0, state.screen_height):
        size = chordpad_width + 1 if ypos == chordpad_height else 1
        yield CursesBlit(
            xpos=vertical_division - 1,
            ypos=ypos,
            text="X" * size,
            color=DEFAULT_COLOR,
            attr=0,
        )


def _update_chord_pad(chord_pad, state):
    chord_pad.clear()
    chord = state.current_indexed_chord().chord
    version = state.get_chord_version(chord)
    version_count = len(state.chord_versions.get(chord, []))
    version_index = state.current_chord_version_index[chord] % version_count + 1
    if version:
        chord_pad.write(0, 0, "{} ({})".format(chord, version), DEFAULT_COLOR)
        chord_pad.write(0, 1, "Version {} of {}".format(version_index,
                                                        version_count),
                        DEFAULT_COLOR)
        lines = draw_chord(version, reverse=state.chord_drawing_is_reversed)
        for line_index, line in enumerate(lines):
            chord_pad.write(0, line_index + 3, line, DEFAULT_COLOR)
            to_paint = ((i, c) for (i, c) in enumerate(line) if c in "xo")
            for index, char in to_paint:
                chord_pad.write(index, line_index + 3, char, FINGERING_COLOR,
                                attr=curses.A_BOLD)
    else:
        chord_pad.write(0, 0, "{}".format(chord), DEFAULT_COLOR)
        chord_pad.write(0, 2, "No fingering available :'(", DEFAULT_COLOR)

def _update_instructions_pad(instructions_pad, dummy_state):
    for line_index, line in enumerate(INSTRUCTIONS):
        instructions_pad.write(0, line_index, line, DEFAULT_COLOR)


def _update_lyrics_pad(lyrics_pad, state):
    lyrics_pad.clear()
    blits = []
    for line_index, line in enumerate(state.tablature.lines):
        current_chord = state.current_indexed_chord()
        selected_chords = ({current_chord.position}
                           if current_chord.line == line_index
                           else {})
        blits.extend(_get_line_blits(line, line_index, state, selected_chords))
    lyrics_pad.write_all(blits)


def _center_at_current_chord(state):
    chord = state.current_indexed_chord()
    return state.center_lyrics_line(chord.line)


def _process_input_key(key, state):
    if key == ord('q'):
        state = state._replace(running=False)
    elif key == curses.KEY_DOWN:
        state = state.move_lyrics(1)
    elif key == curses.KEY_UP:
        state = state.move_lyrics(-1)
    elif key == ord('j'):
        state = state.move_current_chord(-1)
        state = _center_at_current_chord(state)
    elif key == ord('k'):
        state = state.move_current_chord(1)
        state = _center_at_current_chord(state)
    elif key == ord('h'):
        state = state.move_current_chord_version(-1)
    elif key == ord('l'):
        state = state.move_current_chord_version(1)
    elif key == ord('r'):
        state = state.reverse_chord_drawing()
    return state


def _get_line_blits(line, ypos, state, selected_chord_indexes=None):
    selected_chord_indexes = selected_chord_indexes or {}
    res = []
    if line.type == 'lyric':
        res.append(CursesBlit(
            xpos=0,
            ypos=ypos,
            text=line.data.lyrics,
            color=DEFAULT_COLOR,
            attr=0,
        ))
    elif line.type == 'chord':
        position = 0
        for index, (chord_position, chord) in enumerate(line.data.chords):
            attr = (curses.A_STANDOUT | curses.A_BOLD
                    if index in selected_chord_indexes
                    else 0)
            this_pos = max(position, chord_position)
            res.append(CursesBlit(
                xpos=this_pos,
                ypos=ypos,
                text=chord.text(),
                color=CHORD_COLOR,
                attr=attr,
            ))
            position = this_pos + len(chord.text())
            version = state.get_chord_version(chord)
            if version:
                text = "({})".format(version)
                res.append(CursesBlit(
                    xpos=position,
                    ypos=ypos,
                    text=text,
                    color=FINGERING_COLOR,
                    attr=attr,
                ))
                position += len(text)
            position += 1
    return res


INSTRUCTIONS = """
 INSTRUCTIONS/KEYBINDINGS


 k - move to next chord in song
 j - move to previous chord in song
 l - move to next current chord
     version
 h - move to previous current chord
     version
 up/down - navigate
 q - quit
 r - reverse chord orientation
""".strip().splitlines()



'''
Unit tests for utils.regexp
'''

from nose.tools import ok_, raises, eq_

from chorddb.utils.regexp import NoMatchError, re_search, strict


@raises(NoMatchError)
def failed_re_search(pattern, string, flags=0):
    return re_search(pattern, string, flags=flags)


def re_search_test():
    ''' Basic re_search test '''
    regexp = r'a\d'
    ok_(re_search(regexp, "a1"))
    ok_(re_search(regexp, "ba12"))


def invalid_re_search_test():
    ''' Basic invalid re_search test '''
    regexp = r'a\d'
    yield failed_re_search, regexp, "a"
    yield failed_re_search, regexp, "1"
    yield failed_re_search, regexp, "abcdefg"
    yield failed_re_search, regexp, ""


def strict_re_search_test():
    ''' Basic strict re_search test '''
    regexp = r'a\d'
    ok_(re_search(regexp, "a1"))
    ok_(re_search(strict(regexp), "a1"))
    yield failed_re_search, strict(regexp), "ba12"
    yield failed_re_search, strict(regexp), "ba1"
    yield failed_re_search, strict(regexp), "a12"

@raises(ValueError)
def _failed_strict(pattern):
    strict(pattern)

def already_strict_error_test():
    ''' Check for expected error when stricting (semi)strict regexp '''
    for pattern in ['^test', 'test$', '^test$', strict('test')]:
        yield _failed_strict, pattern

def no_match_error_fields_test():
    ''' Test for presense of NoMatchError members '''
    pattern = r'\d{2,3}'
    string = 'this has no numbers'
    try:
        re_search(pattern, string)
    except NoMatchError as err:
        eq_(err.pattern, pattern)
        eq_(err.string, string)





'''
Created on May 21, 2014

@author: ignacio
'''
from .finder import get_fingering_penalty, get_fingerings
from ..utils.decorators import memoize


class ChordLibrary(object):

    def __init__(self, instrument):
        self._instrument = instrument

    def get(self, chord):
        try:
            return self.get_all(chord)[0]
        except IndexError:
            return None

    @memoize
    def get_all(self, chord):
        fingerings = get_fingerings(chord, self._instrument)
        return sorted(fingerings, key=self._get_penalty)

    @staticmethod
    def _get_penalty(fingering):
        return sum(get_fingering_penalty(fingering).values())



#!/usr/bin/env python
# encoding: utf-8

import collections

from .notes import KeyOctave, Key
from .chords.variations import map_variations_to_intervals, VARIATIONS_NOTES


_Instrument = collections.namedtuple('Instrument', [
    'name', 'keyoctaves', 'frets', 'has_bass', 'variation_overrides'], verbose=True)


class Instrument(_Instrument):
    def __new__(cls, *args, **kwargs):
        defaults = {
            'has_bass': False,
            'variation_overrides': {},
        }
        fields_in_args = set(_Instrument._fields[:len(args)])
        kwvalues = {f: kwargs.get(f, defaults[f]) for f in _Instrument._fields
                    if f not in fields_in_args}
        return _Instrument.__new__(cls, *args, **kwvalues)  # pylint: disable=star-args

    @classmethod
    def parse(cls, name, keys, frets, **kwargs):
        keyoctaves = []
        for key in keys:
            if not keyoctaves:
                keyoctaves.append(KeyOctave(key, 0))
            else:
                current = keyoctaves[-1]
                if key <= current.key:
                    keyoctaves.append(KeyOctave(key, current.octave + 1))
                else:
                    keyoctaves.append(KeyOctave(key, current.octave))
        return cls(name, keyoctaves, frets, **kwargs)

    @classmethod
    def from_name(cls, name, default=None):
        try:
            return INSTRUMENTS[name]
        except KeyError:
            if default:
                return default
            raise

    def capo(self, capo_position):
        kwds = {
            'name':"{}(Capo: {})".format(self.name, capo_position),
            'keyoctaves':[ko.transpose(capo_position) for ko in self.keyoctaves],
            'frets':self.frets - capo_position,
        }
        return self._replace(
            name="{}(Capo: {})".format(self.name, capo_position),
            keyoctaves=[ko.transpose(capo_position) for ko in self.keyoctaves],
            frets=self.frets - capo_position,
        )

    def __str__(self):
        return "Instrument: {s.name}".format(s=self)

    def size(self):
        return len(self.keyoctaves)


LOOG_VARIATION_OVERRIDES = {
    '7': map_variations_to_intervals([['G', 'C#', 'E'], ["A", "C#", "G"]]),
    'm7': VARIATIONS_NOTES['m'],
}


GUITAR = Instrument.parse('guitar', [Key(k) for k in list("EADGBE")], 10,
                          has_bass=True)
LOOG = Instrument.parse('loog', [Key(k) for k in list("GBE")], 10,
                        has_bass=False,
                        variation_overrides=LOOG_VARIATION_OVERRIDES)
UKELELE = Instrument('ukelele',
                     [KeyOctave.parse(k) for k in ["G0", "C0", "E0", "A1"]],
                     10, has_bass=False)


INSTRUMENTS = {i.name: i for i in [GUITAR, LOOG, UKELELE]}



'''
Unit tests for utils.decorators module
'''
import itertools

from nose.tools import eq_

from chorddb.utils.decorators import memoize


class CallCounter(object):
    def __init__(self, func):
        self._func = func
        self.__name__ = func.__name__
        self._call_count = 0

    @property
    def call_count(self):
        return self._call_count

    def __call__(self, *args, **kwargs):
        self._call_count += 1
        return self._func(*args, **kwargs)


def noarg_memoize_test():
    ''' Tests memoize on a 0-ary function '''
    value = [1, 2, 3]

    def func():
        return value

    counter = CallCounter(func)
    func = memoize(counter)

    for dummy in xrange(5):
        eq_(func(), value)

    eq_(counter.call_count, 1)


def arg_memoize_test():
    ''' Single arg memoize test '''
    def double(val):
        return 2 * val

    counter = CallCounter(double)
    double = memoize(counter)

    values = [1, 2, 3, 4]
    for val in values:
        for dummy in xrange(5):
            eq_(double(val), 2 * val)
    eq_(counter.call_count, len(values))


def kwarg_memoize_test():
    ''' kwarg memoize test '''
    def multiply(vala, valb):
        return vala * valb

    counter = CallCounter(multiply)
    multiply = memoize(counter)

    values = [1, 2, 3, 4]
    for vala, valb in itertools.product(values, values):
        for dummy in xrange(5):
            eq_(multiply(vala=vala, valb=valb), vala * valb)
            eq_(multiply(valb=valb, vala=vala), vala * valb)
    eq_(counter.call_count, len(values) ** 2)


def unhashable_memoize_test():
    ''' Unhashable arg memoize check '''
    def sort(thing):
        return sorted(thing)

    counter = CallCounter(sort)
    sort = memoize(counter)

    values = [4, 2, 3, 1]
    for dummy in xrange(5):
        eq_(sort(values), sorted(values))


def memoize_cache_immutability_test():
    ''' Test if @memoize cache is immutable '''
    numbers = set([1,2,3,4])
    resp = set(numbers)

    @memoize
    def func():
        return resp

    for _x in xrange(2):
        cached = func()
        eq_(cached, numbers)
        cached.add(5)

    eq_(func(), numbers)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import collections
import logging

from ..chords.library import ChordLibrary


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _get_chord_versions(chords, instrument):
    library = ChordLibrary(instrument)
    return {
        c: library.get_all(c) for c in chords
    }


def _wrap_get(array, index):
    return array[index % len(array)]


IndexedChord = collections.namedtuple('IndexedChord', ['line', 'position',
                                                       'chord'])


def _get_indexed_chords(tablature):
    chords = []
    for line_index, line in enumerate(tablature.lines):
        if line.type == 'chord':
            for chord_index, chord in enumerate(line.data.chords):
                chords.append(IndexedChord(
                    line=line_index, position=chord_index, chord=chord.chord))
    return chords


_RenderState = collections.namedtuple('RenderState', [
    'tablature',
    'instrument',
    'running',
    'screen_width',
    'screen_height',
    'lyrics_position',
    'indexed_chords',
    'current_chord_index',
    'chord_versions',
    'current_chord_version_index',
    'chord_drawing_is_reversed',
])


class RenderState(_RenderState):
    def move_lyrics(self, interval):
        new_position = self.lyrics_position + interval
        return self._replace(
            lyrics_position=self._to_lyrics_position_range(new_position))

    def move_current_chord(self, interval):
        new_chord_index = self.current_chord_index + interval
        return self._replace(current_chord_index=new_chord_index)

    def move_current_chord_version(self, interval):
        current_chord = self.current_indexed_chord().chord
        current_version = dict(self.current_chord_version_index)
        current_version[current_chord] += interval
        return self._replace(current_chord_version_index=current_version)

    def _to_lyrics_position_range(self, lyrics_position):
        # max_position might be negative
        max_position = len(self.tablature.lines) - self.screen_height
        if lyrics_position > max_position:
            lyrics_position = max_position
        if lyrics_position < 0:  # No elif. Fixes the case max_position < 0
            lyrics_position = 0
        return lyrics_position

    def reverse_chord_drawing(self):
        return self._replace(
            chord_drawing_is_reversed=not self.chord_drawing_is_reversed)

    def center_lyrics_line(self, line_index):
        new_position = self._to_lyrics_position_range(
            line_index - self.screen_height / 2)
        return self._replace(lyrics_position=new_position)

    def current_indexed_chord(self):
        return _wrap_get(self.indexed_chords, self.current_chord_index)

    def get_chord_version(self, chord):
        versions = self.chord_versions.get(chord, [])
        return (_wrap_get(versions, self.current_chord_version_index[chord])
                if versions else None)

    @classmethod
    def initial_state(cls, tablature, instrument, screen):
        indexed_chords = _get_indexed_chords(tablature)
        chord_versions = _get_chord_versions(
            set(c.chord for c in indexed_chords), instrument)
        return cls(
            lyrics_position=0,
            current_chord_index=0,
            running=True,
            tablature=tablature,
            instrument=instrument,
            indexed_chords=indexed_chords,
            screen_width=screen.width,
            screen_height=screen.height,
            chord_versions=chord_versions,
            current_chord_version_index=collections.defaultdict(int),
            chord_drawing_is_reversed=False,
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging
from mock import create_autospec, patch, call
from nose.tools import eq_

from chorddb.chords import Chord
from chorddb.tab import transpose_tablature, transpose_line
from chorddb.tab.objects import (
    Tablature, TabLine, ChordLineData, LyricLineData, TabChord)

from test.utils import mock_namedtuple, Sentinels

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@patch('chorddb.tab.transpose_line')
def tanspose_tablature_test(mock_transpose_line):
    one_line = object()
    other_line = object()
    transposed = {
        one_line: object(),
        other_line: object(),
    }
    tab = mock_namedtuple(Tablature, lines=[one_line, other_line])
    mock_transpose_line.side_effect = lambda l, i: transposed[l]
    expected_calls = [call(one_line, 5), call(other_line, 5)]

    res = transpose_tablature(tab, 5)

    eq_(mock_transpose_line.call_args_list, expected_calls)
    eq_(res.lines, [transposed[one_line], transposed[other_line]])


def transpose_lyrics_line_test():
    line = mock_namedtuple(TabLine, type='lyric')
    eq_(transpose_line(line, 5), line)


def transpose_chord_line_test():
    sentinels = Sentinels()
    chord = create_autospec(Chord)
    other_chord = create_autospec(Chord)

    line = mock_namedtuple(
        TabLine,
        type='chord',
        data=mock_namedtuple(
            ChordLineData,
            chords=[mock_namedtuple(TabChord, chord=c)
                    for c in [chord, other_chord]]
        )
    )

    chord.transpose.return_value = sentinels.transposed_chord
    other_chord.transpose.return_value = sentinels.transposed_other_chord

    transposed = transpose_line(line, 5)

    chord.transpose.assert_called_once_with(5)
    other_chord.transpose.assert_called_once_with(5)
    eq_(transposed.type, 'chord')
    new_chords = [c.chord for c in transposed.data.chords]
    eq_(new_chords,
        [sentinels.transposed_chord, sentinels.transposed_other_chord])



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import curses
import logging


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class CursesPad(object):
    def __init__(self, pad, xpos, ypos, width, height, display_width=None,
                 display_height=None):
        self._pad = pad
        self._xpos = xpos
        self._ypos = ypos
        self._width = width
        self._height = height
        self._display_width = display_width or width
        self._display_height = display_height or height
        logger.info("Creating new pad(%s) @%sx%s, size=%sx%s, display=%sx%s",
                    hex(id(self)), xpos, ypos, width, height,
                    self._display_width, self._display_height)

    def write(self, xpos, ypos, text, color, attr=0):
        if not 0 <= ypos < self.height:
            logger.warning("Writing outside pad: ypos=%s not in [0,%s)]",
                           ypos, self.height)
            return
        if xpos >= self.width:
            logger.warning("Writing outside pad: xpos=%s > width=%s",
                           xpos, self.width)
            return
        if xpos < 0:
            logger.warning("Writing outside pad: xpos=%s < 0", xpos)
            text = text[abs(xpos):]
            xpos = 0
        if xpos + len(text) > self.width:
            logger.warning("Writing outside pad: text '%s' is too large "
                           "(len=%s + xpos=%s > width=%s)", text, len(text),
                           xpos, self.width)
            text = text[:self.width - xpos]
        self._write(xpos, ypos, text, color, attr)

    def _write(self, xpos, ypos, text, color, attr=0):
        logger.debug("pad(%s):addstr@%sx%s '%s' color=%s, attr=%s",
                     hex(id(self)), xpos, ypos, text, color, attr)
        self._pad.addstr(ypos, xpos, text, curses.color_pair(color) | attr)

    def write_all(self, blits):
        for blit in blits:
            self.write(blit.xpos, blit.ypos, blit.text, blit.color, blit.attr)

    def clear(self):
        self._pad.clear()

    def refresh(self, vertical_scroll=0):
        self._pad.refresh(vertical_scroll, 0,
                          self._ypos, self._xpos,
                          self._ypos + self._display_height - 1,
                          self._xpos + self._display_width - 1)

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import curses
import logging


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


BOLD = curses.A_BOLD
STANDOUT = curses.A_STANDOUT

DEFAULT_COLOR = 0
CHORD_COLOR = 1
FINGERING_COLOR = 2

_COLORS_DEFINITIONS = {
    CHORD_COLOR: [curses.COLOR_CYAN, curses.COLOR_BLACK],
    FINGERING_COLOR: [curses.COLOR_RED, curses.COLOR_BLACK],
}


def init_colors():
    for color_id, color_pair in _COLORS_DEFINITIONS.items():
        curses.init_pair(color_id, color_pair[0], color_pair[1])



import random
from nose.tools import eq_

from chorddb.chords import Chord
from chorddb.notes import Key

__KEYS = ["A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#"]
_KEYS = [Key.parse(k) for k in __KEYS]
_VARIATIONS = ["m", "7", "m7", "maj7"]


def _check_text_parsing(chord):
    eq_(Chord.parse(chord.text()), chord)


def basic_chord_parse_test():
    """ Parse chords with no variations """
    for key in _KEYS:
        parsed = Chord.parse(key.text())
        eq_(parsed.key, key)
        eq_(parsed.variation, None)
        eq_(parsed.bass, key)
        _check_text_parsing(parsed)


def variation_chord_parse_test():
    """ Parse chords with variations """
    for key in _KEYS:
        for variation in _VARIATIONS:
            parsed = Chord.parse(key.text() + variation)
            eq_(parsed.key, key)
            eq_(parsed.variation, variation)
            eq_(parsed.bass, key)
            _check_text_parsing(parsed)


def bass_chord_parse_test():
    ''' Parse chords with bass '''
    for key in _KEYS:
        for variation in _VARIATIONS + ['']:
            for bass in _KEYS:
                text = key.text() + variation + "/" + bass.text()
                parsed = Chord.parse(text)
                eq_(parsed.key, key)
                eq_(parsed.variation, variation if variation else None)
                eq_(parsed.bass, bass)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging

from nose.tools import ok_


from chorddb.instrument import GUITAR, UKELELE, LOOG
from chorddb.chords.finder import get_fingerings
from chorddb.chords import Chord
from chorddb.notes import Key


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


_TEST_VARIATIONS = [None, "m", "7", "m7"]


def _test_variation_has_fingerings(variation, instrument):
    chord = Chord(Key.parse("C"), variation=variation)
    ok_(len(get_fingerings(chord, instrument)) > 0,
        "Variation has no fingerings! variation={}, instrument={}".format(
            variation, instrument))


def test_variation_have_fingerings():
    for instrument in [GUITAR, UKELELE, LOOG]:
        for variation in _TEST_VARIATIONS:
            yield _test_variation_has_fingerings, variation, instrument





'''
Created on May 17, 2014

@author: ignacio
'''

import collections

from .variations import VARIATIONS_NOTES


PositionedNote = collections.namedtuple('PositionedNote', ['position', 'note'])


def find_fingerings(instrument, keys):
    for positions in search(instrument, keys):
        start = positions.count(None)
        fingering = [x.position for x in positions[start:]]
        yield Fingering(fingering, instrument, start)


BactrackState = collections.namedtuple('BactrackState', ['keys', 'valid_notes',
                                                         'current', 'res'])


def search(instrument, keys):
    return _backtrack(BactrackState(
        keys=set(keys),
        valid_notes=_get_valid_notes(instrument, keys),
        current=[],
        res=[]
    ))


def _backtrack(state):
    if set(c.note.key for c in state.current if c) == state.keys:  # is valid
        state.res.append(list(state.current))  # make a copy
    if len(state.current) == len(state.valid_notes):  # no childs
        return

    used = sorted([c.position for c in state.current if c and c.position])
    maxi = used[-1] if used else 0
    mini = used[0] if used else 10000

    for posnote in state.valid_notes[len(state.current)]:
        if maxi - posnote.position > 4 or posnote.position - mini > 4:
            # Stretch basic test
            continue
        state.current.append(posnote)
        _backtrack(state)
        state.current.pop()

    # Consider empty start:
    if all(c is None for c in state.current):
        state.current.append(None)
        _backtrack(state)
        state.current.pop()

    return state.res

def _get_valid_notes(instrument, keys):
    keys = set(keys)
    valid_notes = []
    for keyoctave in instrument.keyoctaves:
        keyoctaves = (keyoctave.transpose(t) for t in xrange(instrument.frets))
        valid_keyoctaves = (PositionedNote(position=i, note=ko)
                            for i, ko in enumerate(keyoctaves)
                            if ko.key in keys)
        valid_notes.append(list(valid_keyoctaves))
    return valid_notes


def get_fingerings(chord, instrument):
    return list(_get_fingerings(chord, instrument))


def _get_fingerings(chord, instrument):
    if not instrument.has_bass:
        chord = chord.bassless()  # Remove bass if present

    variations = instrument.variation_overrides.get(
        chord.variation, VARIATIONS_NOTES[chord.variation])
    for variation in variations:
        keys = set(chord.key.transpose(v) for v in variation)
        bass_not_in_variation = False
        if instrument.has_bass and not chord.bass in keys:
            keys.add(chord.bass)
            bass_not_in_variation = True

        for fingering in find_fingerings(instrument, keys):
            if instrument.has_bass:
                if fingering.bass().key != chord.bass:
                    continue
                if bass_not_in_variation:
                    bass_count = len([ko for ko in fingering.keyoctaves()
                                      if ko.key == chord.bass])
                    if bass_count > 1:
                        continue
            yield fingering


def _max(fingering):
    return max(x.position for x in fingering if x.position > 0)


def _min(fingering):
    return min(x.position for x in fingering if x.position > 0)


class Fingering(object):
    def __init__(self, positions, instrument, start=None):
        self._instrument = instrument
        self._positions = positions
        self._start = start or 0
        if not 0 <= self._start < instrument.size():
            raise ValueError("Invalid start")
        if instrument.size() < self._start + len(positions):
            raise ValueError("Invalid positions for instrument")
        if not all(0 <= p < instrument.frets for p in positions):
            raise ValueError("Invalid positions for instrument")

    @property
    def positions(self):
        return self._positions

    @property
    def instrument(self):
        return self._instrument

    @property
    def start(self):
        return self._start

    def keyoctaves(self):
        return [ko.transpose(interval) for ko, interval in
                zip(self._instrument.keyoctaves[self._start:],
                    self._positions)]

    def bass(self):
        return min(self.keyoctaves())

    def full_positions(self):
        res = ["x"] * self._instrument.size()
        res[self._start:self._start] = self._positions
        return res[:self._instrument.size()]

    def __str__(self):
        return "".join(str(x) for x in self.full_positions())


def get_fingering_penalty(fingering):
    try:
        bar = min(x for x in fingering.positions if x)
    except ValueError:
        bar = 0

    indexed_poss = sorted(enumerate(x - bar for x in fingering.positions
                                    if x > bar),
                          key=lambda (string, pos): (pos, string),
                          reverse=True)
    fingers = len(indexed_poss)
    if fingers > 4 or (bar and fingers > 3):
        return {'too many fingers': 10000}
    penalty = {
        'start': fingering.start * 5 ** 2,
        'end': (fingering.instrument.size() - fingering.start -
                len(fingering.positions)) * 8 ** 2,
        'positions': sum((p - bar + 2) ** 2 for p in fingering.positions),
        'bar': bar * fingering.instrument.size() * 3 if bar else 0,
        'consecutive_diffs': sum((a - b) ** 2 for a, b in
                                 zip(fingering.positions,
                                     fingering.positions[1:])
                                 if a and b),
        'four_fingers': 50 if fingers == 4 else 0
    }
    return penalty



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging
import re

from ..chords import CHORD_RE, Chord
from .objects import (
    Tablature, TabLine, ChordLineData, LyricLineData, TabChord)


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _parse_chord_line(line):
    ''' Parse a chord line into a `ChordLineData` object. '''
    chords = [
        TabChord(position=position, chord=chord)
        for chord, position in Chord.extract_chordpos(line)
    ]
    return ChordLineData(chords=chords)


def _parse_lyric_line(line):
    ''' Parse a lyric line into a LyricLineData object. '''
    return LyricLineData(lyrics=line)


_DATA_PARSERS = {
    'chord': _parse_chord_line,
    'lyric': _parse_lyric_line,
    'empty': lambda l: None,
}


def _get_line_type(line):
    ''' Decide the line type in function of its contents '''
    stripped = line.strip()
    if not stripped:
        return 'empty'
    remainder = re.sub(r"\s+", " ", re.sub(CHORD_RE, "", stripped))
    if len(remainder) * 2 < len(re.sub(r"\s+", " ", stripped)):
        return 'chord'
    return 'lyric'


def parse_line(line):
    ''' Parse a line into a `TabLine` object. '''
    line = line.rstrip()
    line_type = _get_line_type(line)
    return TabLine(
        type=line_type,
        data=_DATA_PARSERS[line_type](line),
        original=line,
    )


def parse_tablature(lines):
    ''' Parse a list of lines into a `Tablature`. '''
    lines = [parse_line(l) for l in lines]
    return Tablature(lines=lines)



'''
Created on May 25, 2014

@author: ignacio
'''

import itertools
from nose.tools import raises, eq_, ok_

from chorddb.notes import Key, KeyOctave


_NOTES = "ABCDEFG"
_KEYS = ["A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#"]
_KEYOCTAVES = ["{}{}".format(__key, __octave)
               for __octave in xrange(6) for __key in _KEYS]

_INVALID_NOTES = ["H", "a", "b", 1, [], {}, None]
_INVALID_ACCIDENTALS = ['.', '!', 'B', '*', '%']

# Helpers


def _keys():
    ''' Parse all valid keys '''
    return [Key.parse(k) for k in _KEYS]


def _keyoctaves():
    return [KeyOctave.parse(ko) for ko in _KEYOCTAVES]


@raises(ValueError, TypeError)
def invalid_key_parse(key):
    ''' Helper for expenting to fail parses '''
    Key.parse(key)


@raises(ValueError)
def invalid_key_construct(note, accidental=None):
    Key(note, accidental=accidental)


def _check_key_str_parse(key):
    eq_(key, Key.parse(key.text()))


# / Helpers
# Key tests


def note_parse_test():
    ''' Basic note parsing test '''
    for note in _NOTES:
        parsed = Key.parse(note)
        eq_(parsed.note, note)
        eq_(parsed.accidental, None)
        _check_key_str_parse(parsed)


def invalid_key_value_parse_test():
    ''' Check we don't parse invalid Keys '''
    for note in _INVALID_NOTES:
        yield invalid_key_parse, note


def invalid_key_construct_test():
    ''' Check we cant't construct invalid Keys '''
    for note in _INVALID_NOTES:
        yield invalid_key_construct, note
    for accidental in _INVALID_ACCIDENTALS:
        yield invalid_key_construct, "A", accidental


def sharp_parse_test():
    ''' Check sharp parsing '''
    for note in _NOTES:
        parsed = Key.parse(note + "#")
        eq_(Key.parse(note).transpose(1), parsed)
        _check_key_str_parse(parsed)


def flat_parse_test():
    ''' Check flat parsing '''
    for note in _NOTES:
        parsed = Key.parse(note + "b")
        eq_(Key.parse(note).transpose(-1), parsed)
        _check_key_str_parse(parsed)


def key_normalization_test():
    ''' Check equal keys are parsed to equal Keys '''
    for skey, fkey in [('A#', 'Bb'), ('C#', 'Db'), ('D#', 'Eb'),
                       ('F#', 'Gb'), ('G#', 'Ab'), ]:
        return Key.parse(skey) == Key.parse(fkey)


def key_ord_test():
    ''' Check Key.ord/Key.from_ord consistency '''
    for key in _keys():
        eq_(key, Key.from_ord(key.ord()))


def key_eq_test():
    ''' Check different parses of the same key are equal '''
    for key in _KEYS:
        eq_(Key.parse(key), Key.parse(key))


def key_order_test():
    ''' Check Key ordering '''
    for key, okey in itertools.combinations(_keys(), 2):
        ok_(key < okey)
        ok_(key != okey)
        ok_(key.ord() < okey.ord())


def key_transpose_test():
    ''' Check Key transpositions work '''
    keys = _keys()
    for index, key in enumerate(keys):
        for interval in xrange(25):
            nindex = (interval + index) % len(keys)
            eq_(key.transpose(interval), keys[nindex])


# / Key tests
# KeyOctave tests


def keyoctave_parse_test():
    ''' Check valid KeyOctave parses '''
    for key in _KEYS:
        for octave in xrange(10):
            keyoctave = KeyOctave.parse("{}{}".format(key, octave))
            eq_(keyoctave.key, Key.parse(key))
            eq_(keyoctave.octave, octave)


def keyoctave_eq_test():
    ''' Check different parses of the same keyoctave are equal '''
    for keyoctave in _KEYOCTAVES:
        eq_(KeyOctave.parse(keyoctave), KeyOctave.parse(keyoctave))


def keyoctave_ord_test():
    ''' Check Keyoctave's ord/from_ord consistency '''
    for keyoctave in _keyoctaves():
        eq_(keyoctave, KeyOctave.from_ord(keyoctave.ord()))


def keyoctave_order_test():
    ''' Check KeyOctave ordering '''
    for keyoctave, okeyoctave in itertools.combinations(_keyoctaves(), 2):
        ok_(keyoctave < okeyoctave)
        ok_(keyoctave != okeyoctave)
        ok_(keyoctave.ord() < okeyoctave.ord())


def keyoctave_transpose_test():
    ''' Check KeyOctave transpositions work '''
    ikeyoctaves = list(enumerate(_keyoctaves()))
    pairs = itertools.product(ikeyoctaves, ikeyoctaves)

    for (index, keyoctave), (oindex, okeyoctave) in pairs:
        eq_(keyoctave.transpose(oindex - index), okeyoctave)


#  / KeyOctave tests



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import curses
import logging

from .render import render_tablature
from .screen import CursesScreen


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


__all__ = ['wrapper', 'render_tablature']


def wrapper(func):
    def func_wrap(curses_screen):
        screen = CursesScreen(curses_screen)
        try:
            return func(screen)
        except:
            logger.exception("Ooppps!")
            raise
    curses.wrapper(func_wrap)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from StringIO import StringIO
import collections
import logging

from ..chords.library import ChordLibrary

from .color import colorize
from . import color

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


FingeredChord = collections.namedtuple('Fingeredchord', ['chord', 'fingering'])


def _get_fingerings(chords, instrument):
    return [(chord.position,
             FingeredChord(chord=chord.chord,
                           fingering=ChordLibrary(instrument).get(chord.chord)))
            for chord in chords]

def _write_to_buff(buff, text, fore=color.WHITE, back=color.BLACK,
                   style=color.STYLE_NORMAL):
    buff.write(colorize(text, fore, back, style))
    return len(text)

def _render_chord_line(line, instrument):
    chords = _get_fingerings(line.data.chords, instrument)
    buff = StringIO()
    size = 0
    for position, chord in chords:
        if position > size:
            size += _write_to_buff(buff, " " * (position - size))
        _write_to_buff(buff, chord.chord.text(), color.CYAN,
                       style=color.STYLE_BRIGHT)
        if chord.fingering:
            size += _write_to_buff(buff, "({}) ".format(chord.fingering),
                                   color.RED)
    return buff.getvalue()


def _render_line(line, instrument, debug=False):
    if debug:
        print colorize("{:8s}:".format(line.type), color.GREEN,
                       style=color.STYLE_BRIGHT),
    if line.type == 'empty':
        print
    elif line.type == 'lyric':
        print line.data.lyrics
    elif line.type == 'chord':
        print _render_chord_line(line, instrument)
    else:
        print colorize("Unknown line type: '{}'".format(line.type), fore=RED,
                       style=STYLE_BRIGHT)
        print line.original


def render_tablature(tablature, instrument, debug=False):
    for line in tablature.lines:
        _render_line(line, instrument, debug=debug)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import logging

from nose.tools import eq_

from chorddb.tab import parser
from chorddb.chords import Chord

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def parse_chord_line_test():
    parsed = parser.parse_line('A7    Cm    Fm7')
    eq_(parsed.type, 'chord')
    chords = parsed.data.chords
    eq_(len(chords), 3)
    eq_(chords[0].chord, Chord.parse('A7'))
    eq_(chords[0].position, 0)
    eq_(chords[1].chord, Chord.parse('Cm'))
    eq_(chords[1].position, 6)
    eq_(chords[2].chord, Chord.parse('Fm7'))
    eq_(chords[2].position, 12)


def parse_lyric_line_test():
    lyrics = 'This is a lyrics line'
    parsed = parser.parse_line(lyrics)
    eq_(parsed.type, 'lyric')
    eq_(parsed.data.lyrics, lyrics)


def parse_empty_line_test():
    parsed = parser.parse_line('')
    eq_(parsed.type, 'empty')


def parse_tab_test():
    parsed = parser.parse_tablature([
        'A',
        'A line',
        '',
    ])
    lines = parsed.lines
    eq_(len(lines), 3)
    eq_(lines[0].type, 'chord')
    eq_(len(lines[0].data.chords), 1)
    eq_(lines[0].data.chords[0].position, 0)
    eq_(lines[0].data.chords[0].chord, Chord.parse("A"))
    eq_(lines[1].type, 'lyric')
    eq_(lines[1].data.lyrics, 'A line')
    eq_(lines[2].type, 'empty')



'''
Created on May 10, 2014

@author: ignacio
'''
from functools import total_ordering

from .utils.decorators import memoize
from .utils.regexp import strict, re_search

# Some definitions for posterity:
# - Note: ABCDEFG
# - Accidental: # and b
# - Key: note + accidental. e.g Ab
# - variation: major, minor, seventh, maj7, etc

NOTES = list("ABCDEFG")
NOTES_RE = "[{}]".format("".join(NOTES))

ACCIDENTALS = ["b", "#"]
ACCIDENTALS_RE = "[{}]".format("".join(ACCIDENTALS))

KEY_RE = "{}(?:{})?".format(NOTES_RE, ACCIDENTALS_RE)
_CAPTURING_KEY_RE = "({})({})?".format(NOTES_RE, ACCIDENTALS_RE)

KEY_OCTAVE_RE = r"{}\d+".format(KEY_RE)
_CAPTURING_KEY_OCTAVE_RE = r"({})(\d+)".format(KEY_RE)

_NOTE_ACCIDENTALS = {
    "A": "b#",
    "B": "b",
    "C": "#",
    "D": "b#",
    "E": "b",
    "F": "#",
    "G": "b#",
}

NOTE_ACCIDENTALS = dict((k, list(v)) for (k, v) in _NOTE_ACCIDENTALS.items())
KEY_NORMALIZATIONS = {
    "Ab": "G#",
    "Bb": "A#",
    "B#": "C",
    "Cb": "B",
    "Db": "C#",
    "Eb": "D#",
    "E#": "F",
    "Fb": "E",
    "Gb": "F#",
}


@total_ordering
class Key(object):

    def __init__(self, note, accidental=None):
        if note not in NOTES:
            raise ValueError("Invalid note: '{}'".format(note))
        if accidental and accidental not in ACCIDENTALS:
            raise ValueError("Invalid accidental: '{}'".format(accidental))
        note, accidental = _normalize(note, accidental)
        self._note = note
        self._accidental = accidental

    @property
    def note(self):
        return self._note

    @property
    def accidental(self):
        return self._accidental

    def ord(self):
        return self.all().index(self)

    def transpose(self, interval):
        index = (self.all().index(self) + interval) % len(self.all())
        return self.all()[index]

    def __eq__(self, okey):
        return self.note == okey.note and self.accidental == okey.accidental

    def __ne__(self, okey):
        return not self == okey

    def __lt__(self, okey):
        return (self.note < okey.note or
                (self.note == okey.note and
                 self._accidental_lt(self.accidental, okey.accidental)))

    def __str__(self):
        return "Key:{}".format(self.text())

    def text(self):
        return self.note + self.accidental if self.accidental else self.note

    def __hash__(self):
        return hash((self._note, self._accidental,))

    @classmethod
    def parse(cls, key):
        note, accidental = re_search(strict(_CAPTURING_KEY_RE), key).groups()
        return cls(note, accidental)

    @classmethod
    def from_ord(cls, index):
        return cls.all()[index]

    @classmethod
    @memoize
    def all(cls):
        res = set()
        for note, accidentals in NOTE_ACCIDENTALS.items():
            for accidental in accidentals:
                res.add(cls(note, accidental))
            res.add(cls(note))
        return sorted(res)

    @staticmethod
    def _accidental_lt(accidental, oaccidental):
        if accidental == 'b':
            return oaccidental is None or oaccidental == '#'
        elif accidental is None:
            return oaccidental == "#"
        return False


@total_ordering
class KeyOctave(object):

    def __init__(self, key, octave):
        self._key = key
        self._octave = octave

    @property
    def key(self):
        return self._key

    @property
    def octave(self):
        return self._octave

    def transpose(self, interval):
        return self.from_ord(self.ord() + interval)

    @classmethod
    def from_ord(cls, value):
        keys = Key.all()
        octave, index = divmod(value, len(keys))
        return cls(keys[index], octave)

    def ord(self):
        keys = Key.all()
        return self._octave * len(keys) + keys.index(self._key)

    def __eq__(self, okey):
        return self.key == okey.key and self.octave == okey.octave

    def __ne__(self, okey):
        return not self == okey

    def __lt__(self, okey):
        return (self.octave < okey.octave or
                (self.octave == okey.octave and self.key < okey.key))

    def __str__(self):
        return "KeyOctave:{}".format(self.text())

    def text(self):
        return "{}{}".format(self._key.text(), self._octave)

    def __hash__(self):
        return hash((self._key, self._octave,))

    @classmethod
    def parse(cls, key_octave):
        key, octave = re_search(strict(_CAPTURING_KEY_OCTAVE_RE),
                                key_octave).groups()
        return cls(Key.parse(key), int(octave))


def _normalize(note, accidental):
    if accidental:
        key = note + accidental
        nkey = KEY_NORMALIZATIONS.get(key, key)
        note, accidental = nkey if len(nkey) > 1 else (nkey, None)
    return note, accidental
